import platform

from sqlalchemy.testing import exclusions, fails_if
from sqlalchemy.testing.exclusions import LambdaPredicate, skip_if, skip
from sqlalchemy.testing.requirements import SuiteRequirements


class Requirements(SuiteRequirements):
    @property
    def implicitly_named_constraints(self):
        return exclusions.open()

    @property
    def temp_table_reflection(self):
        return exclusions.closed()

    @property
    def ctes(self):
        return exclusions.open()

    @property
    def autocommit(self):
        return exclusions.open()

    @property
    def reflects_pk_names(self):
        return exclusions.open()

    @property
    def datetime(self):
        return exclusions.closed()

    @property
    def datetime_microseconds(self):
        return exclusions.closed()

    @property
    def empty_inserts(self):
        return exclusions.closed()

    @property
    def implicit_decimal_binds(self):
        return exclusions.closed()

    @property
    def unbounded_varchar(self):
        return exclusions.closed()

    @property
    def time_microseconds(self):
        return exclusions.closed()



#!/usr/bin/env python

"""Tests for `iaccess` package."""

import pytest
from sqlalchemy.dialects import registry

# noinspection PyUnresolvedReferences
from iaccess.dialect import IAccessDialect

registry.register("iaccess.pyodbc", "iaccess.dialect", "IAccessDialect")

pytest.register_assert_rewrite("sqlalchemy.testing.assertions")

# noinspection PyUnresolvedReferences
from sqlalchemy.testing.plugin.pytestplugin import *



# test/test_suite.py

# noinspection PyUnresolvedReferences
import pytest
from sqlalchemy.testing.suite import *


class CTETest(CTETest):
    @pytest.mark.xfail(reason="ctes not allowed in insert statements on db2 for iseries")
    def test_insert_from_select_round_trip(self):
        super().test_insert_from_select_round_trip()


class ComponentReflectionTest(ComponentReflectionTest):
    @pytest.mark.xfail(reason="unique constraints with duplicate column sets unsupported on db2 for iseries")
    def test_get_unique_constraints(self):
        return super().test_get_unique_constraints()

    @pytest.mark.xfail(reason="unknown problem with test -- deprecation error not raised")
    def test_deprecated_get_primary_keys(self):
        return super().test_deprecated_get_primary_keys


class ExpandingBoundInTest(ExpandingBoundInTest):
    @pytest.mark.xfail(reason="search condition where null in set not unsupported on db2 for iseries")
    def test_null_in_empty_set_is_false(self):
        return super().test_null_in_empty_set_is_false()


class InsertBehaviorTest(InsertBehaviorTest):
    @pytest.mark.xfail(reason="iaccess odbc driver is not able to infer some data types")
    def test_insert_from_select_with_defaults(self):
        return super().test_insert_from_select_with_defaults()


class NumericTest(NumericTest):
    @pytest.mark.xfail(reason="iaccess odbc driver is not able to infer some data types")
    def test_float_coerce_round_trip(self):
        return super().test_float_coerce_round_trip()


class TableDDLTest(TableDDLTest):
    @pytest.mark.xfail(reason="need to specify test schema")
    def test_create_table_schema(self):
        return super().test_create_table_schema()


class StringTest(StringTest):
    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal_non_ascii(self):
        super().test_literal_non_ascii()


class TextTest(TextTest):
    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal_non_ascii(self):
        super().test_literal_non_ascii()


class UnicodeVarcharTest(UnicodeVarcharTest):
    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal_non_ascii(self):
        super().test_literal_non_ascii()

    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal(self):
        super().test_literal()


class UnicodeTextTest(UnicodeTextTest):
    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal_non_ascii(self):
        super().test_literal_non_ascii()
        
    @pytest.mark.skip(reason="high unicode literals cannot fit in UCS-2 encoding used by iaccess driver")
    def test_literal(self):
        super().test_literal()



import re

from sqlalchemy.sql import compiler

ILLEGAL_INITIAL_CHARACTERS = {str(x) for x in range(0, 10)}.union(["$", "_"])

_MAX_BIGINT = 2**63 - 1  # 9223372036854775807


class IAccessCompiler(compiler.SQLCompiler):
    @staticmethod
    def default_from():
        return " FROM SYSIBM.SYSDUMMY1"

    def visit_empty_set_expr(self, element_types):
        # noinspection SqlConstantCondition
        return ' '.join(['SELECT 1', self.default_from(), 'WHERE 1!=1'])

    def limit_clause(self, select, **kw):
        text = ""
        if select._limit_clause is not None:
            text += "\n LIMIT " + self.process(select._limit_clause, **kw)
        if select._offset_clause is not None:
            if select._limit_clause is None:
                text += "\n LIMIT " + str(_MAX_BIGINT)
            text += " OFFSET " + self.process(select._offset_clause, **kw)
        return text

    def visit_sequence(self, sequence, **kw):
        seq = self.dialect.identifier_preparer.format_sequence(sequence)
        return f"NEXT VALUE FOR {seq}"


class IAccessDDLCompiler(compiler.DDLCompiler):
    def visit_create_table(self, create):
        ddl = super().visit_create_table(create)

        # IBM supports `DECLARE GLOBAL TEMPORARY TABLE` instead of `CREATE GLOBAL TEMPORARY TABLE`
        if re.match('\W*CREATE GLOBAL TEMPORARY', ddl):
            ddl = ddl.replace('CREATE GLOBAL', 'DECLARE GLOBAL', 1)
        return ddl

    def get_column_specification(self, column, **kwargs):
        colspec = (
            self.preparer.format_column(column)
            + " "
            + self.dialect.type_compiler.process(
                column.type, type_expression=column
            )
        )

        system_name = kwargs.get('system_name', None)
        if system_name is not None:
            colspec += " FOR COLUMN " + system_name

        default = self.get_column_default_string(column)
        if default is not None:
            colspec += " DEFAULT " + default

        if column.computed is not None:
            colspec += " " + self.process(column.computed)

        if (
            column.table is not None
            and column is column.table._autoincrement_column
            and column.server_default is None
        ):
            colspec += " GENERATED BY DEFAULT AS IDENTITY "

        implicitly_hidden = kwargs.get('implicitly_hidden', None)
        if implicitly_hidden is not None:
            colspec += " IMPLICITLY HIDDEN"

        if not column.nullable:
            colspec += " NOT NULL"
        return colspec


class IAccessTypeCompiler(compiler.GenericTypeCompiler):
    def visit_BOOLEAN(self, type_, **kw):
        return self.visit_SMALLINT(type_, **kw)

    def visit_DATETIME(self, type_, **kw):
        return self.visit_TIMESTAMP(type_, **kw)

    def visit_TEXT(self, type_, **kw):
        return self.visit_CLOB(type_, **kw)

    def visit_unicode(self, type_, **kw):
        return self.visit_NVARCHAR(type_, **kw)

    def visit_unicode_text(self, type_, **kw):
        return self.visit_NCLOB(type_, **kw)


class IAccessIdentifierPreparer(compiler.IdentifierPreparer):
    illegal_initial_characters = ILLEGAL_INITIAL_CHARACTERS



from sqlalchemy import MetaData, Table, Column, String, Integer

ischema = MetaData()

schema = Table(
    'SYSSCHEMAS',
    ischema,
    Column('SCHEMA_NAME', String, key='schema_name'),
    schema='QSYS2',
)

tables = Table(
    'SYSTABLES',
    ischema,
    Column('TABLE_NAME', String, key='table_name'),
    Column('TABLE_SCHEMA', String, key='table_schema'),
    Column('TABLE_TYPE', String, key='table_type'),
    schema='QSYS2',
)

columns = Table(
    'SYSCOLUMNS',
    ischema,
    Column('COLUMN_NAME', String, key='column_name'),
    Column('DATA_TYPE', String, key='data_type'),
    Column('LENGTH', String, key='length'),
    Column('NUMERIC_SCALE', String, key='numeric_scale'),
    Column('NUMERIC_PRECISION', String, key='numeric_precision'),
    Column('TABLE_NAME', String, key='table_name'),
    Column('TABLE_SCHEMA', String, key='table_schema'),
    Column('COLUMN_DEFAULT', String, key='column_default'),
    Column('IS_NULLABLE', String, key='is_nullable'),
    Column('IS_IDENTITY', String, key='is_identity'),
    Column('IDENTITY_GENERATION', String, key='identity_generation'),
    Column('IDENTITY_START', String, key='identity_start'),
    Column('IDENTITY_INCREMENT', String, key='identity_increment'),
    Column('IDENTITY_MINIMUM', String, key='identity_minimum'),
    Column('IDENTITY_MAXIMUM', String, key='identity_maximum'),
    Column('IDENTITY_CYCLE', String, key='identity_cycle'),
    Column('IDENTITY_CACHE', String, key='identity_cache'),
    Column('IDENTITY_ORDER', String, key='identity_order'),
    schema='QSYS2',
)

indexes = Table(
    'SYSINDEXES',
    ischema,
    Column('INDEX_NAME', String, key='index_name'),
    Column('TABLE_NAME', String, key='table_name'),
    Column('TABLE_SCHEMA', String, key='table_schema'),
    Column('COLUMN_COUNT', String, key='column_count'),
    Column('IS_UNIQUE', String, key='is_unique'),
    Column('INDEX_SCHEMA', String, key='index_schema'),
    Column('INDEX_SCHEMA', String, key='index_schema'),
    schema='QSYS2',
)

keys = Table(
    'SYSKEYS',
    ischema,
    Column('INDEX_NAME', String, key='index_name'),
    Column('COLUMN_NAME', String, key='column_name'),
    Column('ORDINAL_POSITION', Integer, key='ordinal_position'),
    Column('COLUMN_NAME', String, key='column_name'),
    schema='QSYS2',
)

constraints = Table(
    'SYSCST',
    ischema,
    Column('CONSTRAINT_NAME', String, key='constraint_name'),
    Column('TABLE_NAME', String, key='table_name'),
    Column('TABLE_SCHEMA', String, key='table_schema'),
    Column('CONSTRAINT_TYPE', String, key='constraint_type'),
)


constraint_columns = Table(
    'SYSCSTCOL',
    ischema,
    Column('CONSTRAINT_NAME', String, key='constraint_name'),
    Column('TABLE_NAME', String, key='table_name'),
    Column('TABLE_SCHEMA', String, key='table_schema'),
    Column('COLUMN_NAME', String, key='column_name'),
)


sequences = Table(
    'SYSSEQUENCES',
    ischema,
    Column('SEQUENCE_SCHEMA', String, key='sequence_schema'),
    Column('SEQUENCE_NAME', String, key='sequence_name'),
    schema='QSYS2'
)



from sqlalchemy.testing.provision import temp_table_keyword_args


@temp_table_keyword_args.for_db("iaccess")
def _iseries_temp_table_keyword_args(cfg, eng):
    return {
        "prefixes": ['GLOBAL TEMPORARY'],
    }



from itertools import groupby

from iaccess.compiler import IAccessCompiler, IAccessDDLCompiler, IAccessTypeCompiler, IAccessIdentifierPreparer
from iaccess.connector import IAccessConnector
from iaccess.information_schema import iseries as ischema
from sqlalchemy import sql, util, and_
from sqlalchemy.connectors.pyodbc import PyODBCConnector
from sqlalchemy.engine import default, reflection
from sqlalchemy.sql import sqltypes
from sqlalchemy.types import BLOB, DATE, DATETIME, SMALLINT, BIGINT, INTEGER, CHAR, VARCHAR, CLOB, DECIMAL, NUMERIC, \
    REAL, FLOAT, TIME, TIMESTAMP


class DOUBLE(sqltypes.Numeric):
    __visit_name__ = 'DOUBLE'


class LONGVARCHAR(sqltypes.VARCHAR):
    __visit_name_ = 'LONGVARCHAR'


class DBCLOB(sqltypes.CLOB):
    __visit_name__ = "DBCLOB"


class GRAPHIC(sqltypes.CHAR):
    __visit_name__ = "GRAPHIC"


class VARGRAPHIC(sqltypes.Unicode):
    __visit_name__ = "VARGRAPHIC"


class LONGVARGRAPHIC(sqltypes.UnicodeText):
    __visit_name__ = "LONGVARGRAPHIC"


class XML(sqltypes.Text):
    __visit_name__ = "XML"


ischema_names = {
    'BLOB': BLOB,
    'CHAR': CHAR,
    'CHARACTER': CHAR,
    'CLOB': CLOB,
    'DATE': DATE,
    'DATETIME': TIMESTAMP,
    'INTEGER': INTEGER,
    'SMALLINT': SMALLINT,
    'BIGINT': BIGINT,
    'DECIMAL': DECIMAL,
    'NUMERIC': NUMERIC,
    'REAL': REAL,
    'DOUBLE': DOUBLE,
    'FLOAT': FLOAT,
    'TIME': TIME,
    'TIMESTAMP': TIMESTAMP,
    'VARCHAR': VARCHAR,
    'LONGVARCHAR': LONGVARCHAR,
    'XML': XML,
    'GRAPHIC': GRAPHIC,
    'VARGRAPHIC': VARGRAPHIC,
    'LONGVARGRAPHIC': LONGVARGRAPHIC,
    'DBCLOB': DBCLOB,
    'TIMESTMP': TIMESTAMP,
}


class IAccessExecutionContext(default.DefaultExecutionContext):
    def get_lastrowid(self):
        s = sql.select([sql.literal_column('IDENTITY_VAL_LOCAL()')])
        return self.connection.scalar(s)

    def fire_sequence(self, sequence, type_):
        d: IAccessDialect = self.dialect
        return self._execute_scalar(
            "SELECT NEXT VALUE FOR "
            + d.identifier_preparer.format_sequence(sequence)
            + d.statement_compiler.default_from(),
            type_,
        )


class IAccessDialect(IAccessConnector, default.DefaultDialect):
    name = 'iaccess'
    encoding = 'utf-8'
    default_param_style = 'qmark'
    schema_name = 'QGPL'

    max_identifier_length = 128  # https://www.ibm.com/support/knowledgecenter/en/ssw_ibm_i_74/IAccess/rbafzlimtabs.htm
    supports_alter = True

    supports_sane_rowcount = False
    supports_sane_rowcount_returning = False
    supports_native_boolean = False
    supports_sequences = True
    sequences_optional = True

    supports_default_values = False

    requires_name_normalize = True

    two_phase_transactions = False

    statement_compiler = IAccessCompiler
    ddl_compiler = IAccessDDLCompiler
    type_compiler = IAccessTypeCompiler
    preparer = IAccessIdentifierPreparer
    execution_ctx_cls = IAccessExecutionContext

    @classmethod
    def dbapi(cls):
        import pyodbc
        return pyodbc

    def has_table(self, connection, table_name, schema=None):
        tables = ischema.tables
        table_name = self.denormalize_name(table_name)
        schema = self.denormalize_name(schema)

        whereclause = sql.and_(
            tables.c.table_name == table_name,
            tables.c.table_type == 'T',
            tables.c.table_schema == sql.func.coalesce(schema, sql.literal_column('CURRENT_SCHEMA'))
        )

        s = sql.select([tables], whereclause)
        c = connection.execute(s)
        return c.first() is not None

    def _get_default_schema_name(self, connection):
        query = sql.text("select CURRENT_SCHEMA from sysibm.sysdummy1")
        default_schema_name = connection.scalar(query)
        if default_schema_name is not None:
            # guard against the case where the default_schema_name is being
            # fed back into a table reflection function.
            return sql.quoted_name(default_schema_name, quote=True)
        else:
            return self.schema_name

    @reflection.cache
    def get_table_names(self, connection, schema=None, **kw):
        tables = ischema.tables
        schema = self.denormalize_name(schema or self.default_schema_name)
        s = sql.select(
            [tables.c.table_name],
            sql.and_(
                tables.c.table_schema == schema,
                tables.c.table_type == 'T',
            ),
            order_by=[tables.c.table_name],
        )
        table_names = [self.normalize_name(r.table_name) for r in connection.execute(s)]
        return table_names

    @reflection.cache
    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        requested_schema = schema
        table_name = self.denormalize_name(table_name)
        schema = self.denormalize_name(schema or self.default_schema_name)
        s = """
        select fk.table_name, cst.constraint_name, fk.column_name as constrained_column,
               tgt.column_name as referred_column, tgt.table_name as referred_table,
               tgt.table_schema as referred_schema
            from qsys2.syscst cst
                     join qsys2.syskeycst fk
                          on cst.constraint_schema = fk.constraint_schema
                              and cst.constraint_name = fk.constraint_name
                     join qsys2.sysrefcst ref
                          on cst.constraint_schema = ref.constraint_schema
                              and cst.constraint_name = ref.constraint_name
                     join qsys2.syskeycst pk
                          on ref.unique_constraint_schema = pk.constraint_schema
                              and ref.unique_constraint_name = pk.constraint_name
                     join qsys2.syscolumns tgt
                          on pk.table_schema = tgt.table_schema
                              and pk.table_name = tgt.table_name
                              and pk.column_position = tgt.ordinal_position
            where cst.constraint_type = 'FOREIGN KEY'
              and fk.ordinal_position = pk.ordinal_position
              and fk.table_schema = COALESCE(?, CURRENT_SCHEMA)
              and fk.table_name = ?
              and enabled = 'YES'"""
        r = connection.execute(s, [schema, table_name])
        results = []
        for fk in r:
            referred_schema = requested_schema
            if requested_schema is not None:
                referred_schema = self.normalize_name(fk.referred_schema)
            results.append({
                'name': self.normalize_name(fk.constraint_name),
                'constrained_columns': [self.normalize_name(fk.constrained_column)],
                'referred_schema': referred_schema,
                'referred_table': self.normalize_name(fk.referred_table),
                'referred_columns': [self.normalize_name(fk.referred_column)],
            })
        return results

    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kw):
        current_schema = self.denormalize_name(schema or self.default_schema_name)
        table_name = self.denormalize_name(table_name)
        columns = ischema.columns
        s = sql.select([
            columns.c.column_name,
            columns.c.data_type,
            columns.c.length,
            columns.c.numeric_precision,
            columns.c.numeric_scale,
            columns.c.is_nullable,
            columns.c.column_default,
            columns.c.is_identity,
            columns.c.identity_generation,
            columns.c.identity_start,
            columns.c.identity_increment,
            columns.c.identity_minimum,
            columns.c.identity_maximum,
            columns.c.identity_cycle,
            columns.c.identity_cache,
            columns.c.identity_order,
        ], sql.and_(
            columns.c.table_name == table_name,
            columns.c.table_schema == current_schema,
        ))
        r = connection.execute(s)

        results = []
        for col in r:
            data_type = ischema_names.get(col.data_type, None)
            if data_type is None:
                util.warn(
                    "Did not recognize type '%s' of column '%s'"
                    % (col.data_type, col.column_name)
                )
                data_type = sqltypes.NULLTYPE

            if issubclass(data_type, sqltypes.Numeric) and not issubclass(data_type, sqltypes.Float):
                data_type = data_type(precision=col.numeric_precision,
                                      scale=col.numeric_scale)
            elif issubclass(data_type, sqltypes.String) and col.length:
                data_type = data_type(col.length)

            additional = {}

            auto_increment = col.is_identity and col.identity_generation is not None
            if auto_increment:
                additional['sequence'] = {
                    'name': '',
                    'start': col.identity_start,
                    'increment': col.identity_increment,
                    'minvalue': col.identity_minimum,
                    'maxvalue': col.identity_maximum,
                    'nominvalue': False,
                    'nomaxvalue': False,
                    'cycle': col.identity_cycle,
                    'cache': col.identity_cache,
                    'order': col.identity_order
                }

            col_dict = {
                'name': self.normalize_name(col.column_name),
                'type': data_type,
                'nullable': col.is_nullable == 'Y',
                'default': col.column_default,
                'autoincrement': auto_increment,
            }
            col_dict.update(additional)
            results.append(col_dict)
        return results

    @reflection.cache
    def get_indexes(self, connection, table_name, schema=None, **kw):
        table_name = self.denormalize_name(table_name)
        schema = self.denormalize_name(schema or self.default_schema_name)
        indexes = ischema.indexes
        keys = ischema.keys
        j = indexes.join(keys, indexes.c.index_name == keys.c.index_name)
        s = sql.select([
            indexes.c.index_name,
            indexes.c.is_unique,
            keys.c.column_name,
        ], and_(
            indexes.c.table_schema == schema,
            indexes.c.table_name == table_name,
        )).select_from(j).order_by(indexes.c.index_name, keys.c.ordinal_position)
        r = connection.execute(s)
        grouped = groupby(r, lambda x: (x.index_name, x.is_unique in 'VU'))
        results = [dict(name=self.normalize_name(index_name),
                        column_names=[self.normalize_name(c.column_name) for c in columns],
                        unique=is_unique,
                        ) for (index_name, is_unique), columns in grouped]
        return results

    @reflection.cache
    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        table_name = self.denormalize_name(table_name)
        schema = self.denormalize_name(schema or self.default_schema_name)
        constraints = ischema.constraints
        constraint_columns = ischema.constraint_columns
        j = constraints.join(constraint_columns, and_(
            constraints.c.table_schema == constraint_columns.c.table_schema,
            constraints.c.table_name == constraint_columns.c.table_name,
            constraints.c.constraint_name == constraint_columns.c.constraint_name,
        ))
        s = sql.select([
            constraint_columns.c.constraint_name,
            constraint_columns.c.column_name,
        ], and_(
            constraints.c.constraint_type == 'PRIMARY KEY',
            constraints.c.table_name == table_name,
            constraints.c.table_schema == schema,
        )).select_from(j)
        r = connection.execute(s)
        grouped = groupby(r, lambda x: self.normalize_name(x.constraint_name))
        results = [{
            'constrained_columns': [self.normalize_name(col.column_name) for col in columns],
            'name': cst_name,
        } for (cst_name, columns) in grouped]
        result = results[0] if results else None
        return result

    def has_sequence(self, connection, sequence_name, schema=None):
        schema = self.denormalize_name(schema or self.default_schema_name)
        sequence_name = self.denormalize_name(sequence_name)
        sequences = ischema.sequences
        s = sql.select(
            [sequences.c.sequence_name],
            and_(
                sequences.c.sequence_name == sequence_name,
                sequences.c.sequence_schema == schema,
            ))
        r = connection.execute(s)
        return r.first() is not None



#!/usr/bin/env python

"""The setup script."""
import re
from pathlib import Path

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

version_file = Path('iaccess/__init__.py')
version = re.match(
    r""".*__version__ = '(.*?)'""", version_file.read_text(), re.S
).group(1)

requirements = ['sqlalchemy', 'pyodbc<4.0.30']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Steven Clayton James",
    author_email='steven@waitforitjames.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Provides support for DB2 for iSeries for remote python clients "
                "using IBM i Access Client Solutions ODBC driver",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='iaccess',
    name='iaccess',
    packages=find_packages(include=['iaccess', 'iaccess.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/soundstripe/iaccess',
    version=version,
    zip_safe=False,
    entry_points={
     'sqlalchemy.dialects': [
          'iaccess.pyodbc = iaccess.dialect:IAccessDialect',
          'iaccess = iaccess.dialect:IAccessDialect',
          ]
    }
)



import re
import warnings

import pyodbc
from sqlalchemy.connectors.pyodbc import PyODBCConnector

CLIENT_SOLUTIONS_URL = 'https://www.ibm.com/support/pages/odbc-driver-ibm-i-access-client-solutions'


class IAccessDriverMissingWarning(UserWarning):
    pass


class IAccessConnector(PyODBCConnector):
    pyodbc_driver_name = 'IBM i Access ODBC Driver'

    def create_connect_args(self, url):
        available_drivers = pyodbc.drivers()
        selected_driver = url.query.get('DRIVER', self.pyodbc_driver_name)
        if selected_driver not in available_drivers:
            warnings.warn(f'driver "{selected_driver}" not found by pyodbc\n'
                          f'install from {CLIENT_SOLUTIONS_URL} or specify an alternate driver name\n'
                          f'available drivers: {repr(available_drivers or None)}',
                          IAccessDriverMissingWarning)

        connectors, connect_args = super().create_connect_args(url)

        # force unicode mode (utf-16 instead of ebdic default)
        connect_args.update(unicodesql=1)

        # set true autocommit on (without this, autocommit forces isolation mode *NONE
        connect_args.update(trueautocommit=1)

        # non-standard "system" keyword required instead of "server"
        server = re.search('server=([^;]+)', connectors[0], re.IGNORECASE)
        if server:
            connect_args.update(system=server.group(1))

        return [connectors, connect_args]
